##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
# Example objects for pickling.

from persistent import Persistent

def print_dict(d):
    d = sorted(d.items())
    print('{%s}' % (', '.join(
        [('%r: %r' % (k, v)) for (k, v) in d]
        )))

def cmpattrs(self, other, *attrs):
    result = 0
    for attr in attrs:
        if attr[:3] in ('_v_', '_p_'):
            raise AssertionError("_v_ and _p_ attrs not allowed")
        lhs = getattr(self, attr, None)
        rhs = getattr(other, attr, None)
        result += lhs != rhs
    return result

class Simple(Persistent):
    def __init__(self, name, **kw):
        self.__name__ = name
        self.__dict__.update(kw)
        self._v_favorite_color = 'blue'
        self._p_foo = 'bar'

    @property
    def _attrs(self):
        return list(self.__dict__.keys())

    def __eq__(self, other):
        return cmpattrs(self, other, '__class__', *self._attrs) == 0


class Custom(Simple):

    def __new__(cls, x, y):
        r = Persistent.__new__(cls)
        r.x, r.y = x, y
        return r

    def __init__(self, x, y):
        self.a = 42

    def __getnewargs__(self):
        return self.x, self.y

    def __getstate__(self):
        return self.a

    def __setstate__(self, a):
        self.a = a


class Slotted(Persistent):

    __slots__ = 's1', 's2', '_p_splat', '_v_eek'

    def __init__(self, s1, s2):
        self.s1, self.s2 = s1, s2
        self._v_eek = 1
        self._p_splat = 2

    @property
    def _attrs(self):
        raise NotImplementedError()

    def __eq__(self, other):
        return cmpattrs(self, other, '__class__', *self._attrs) == 0


class SubSlotted(Slotted):

    __slots__ = 's3', 's4'

    def __init__(self, s1, s2, s3):
        Slotted.__init__(self, s1, s2)
        self.s3 = s3

    @property
    def _attrs(self):
        return ('s1', 's2', 's3', 's4')


class SubSubSlotted(SubSlotted):

    def __init__(self, s1, s2, s3, **kw):
        SubSlotted.__init__(self, s1, s2, s3)
        self.__dict__.update(kw)
        self._v_favorite_color = 'blue'
        self._p_foo = 'bar'

    @property
    def _attrs(self):
        return ['s1', 's2', 's3', 's4'] + list(self.__dict__.keys())



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Persistence Interfaces
"""

from zope.interface import Interface
from zope.interface import Attribute

from persistent._compat import use_c_impl


# Allowed values for _p_state. Use the C constants if available (which
# are defined in cPersistence --- there is no cInterfaces --- so we
# need to pass the corresponding ``mod_name``), otherwise define some
# values here.
GHOST = use_c_impl(-1, 'GHOST', mod_name='persistent.persistence')
UPTODATE = use_c_impl(0, 'UPTODATE', mod_name='persistent.persistence')
CHANGED = use_c_impl(1, 'CHANGED', mod_name='persistent.persistence')
STICKY = use_c_impl(2, 'STICKY', mod_name='persistent.persistence')


OID_TYPE = SERIAL_TYPE = bytes

class IPersistent(Interface):
    """Python persistent interface

    A persistent object can be in one of several states:

    - Unsaved

      The object has been created but not saved in a data manager.

      In this state, the _p_changed attribute is non-None and false
      and the _p_jar attribute is None.

    - Saved

      The object has been saved and has not been changed since it was saved.

      In this state, the _p_changed attribute is non-None and false
      and the _p_jar attribute is set to a data manager.

    - Sticky

      This state is identical to the saved state except that the
      object cannot transition to the ghost state.  This is a special
      state used by C methods of persistent objects to make sure that
      state is not unloaded in the middle of computation.

      In this state, the _p_changed attribute is non-None and false
      and the _p_jar attribute is set to a data manager.

      There is no Python API for detecting whether an object is in the
      sticky state.

    - Changed

      The object has been changed.

      In this state, the _p_changed attribute is true
      and the _p_jar attribute is set to a data manager.

    - Ghost

      the object is in memory but its state has not been loaded from
      the database (or its state has been unloaded).  In this state,
      the object doesn't contain any application data.

      In this state, the _p_changed attribute is None, and the _p_jar
      attribute is set to the data manager from which the object was
      obtained.

    In all the above, _p_oid (the persistent object id) is set when
    _p_jar first gets set.

    The following state transitions are possible:

    - Unsaved -> Saved

      This transition occurs when an object is saved in the
      database.  This usually happens when an unsaved object is added
      to (e.g. as an attribute or item of) a saved (or changed) object
      and the transaction is committed.

    - Saved  -> Changed
      Sticky -> Changed
      Ghost  -> Changed

      This transition occurs when someone sets an attribute or sets
      _p_changed to a true value on a saved, sticky or ghost object.  When
      the transition occurs, the persistent object is required to call the
      register() method on its data manager, passing itself as the
      only argument.

      Prior to ZODB 3.6, setting _p_changed to a true value on a ghost object
      was ignored (the object remained a ghost, and getting its _p_changed
      attribute continued to return None).

    - Saved -> Sticky

      This transition occurs when C code marks the object as sticky to
      prevent its deactivation.

    - Saved -> Ghost

      This transition occurs when a saved object is deactivated or
      invalidated.  See discussion below.

    - Sticky -> Saved

      This transition occurs when C code unmarks the object as sticky to
      allow its deactivation.

    - Changed -> Saved

      This transition occurs when a transaction is committed.  After
      saving the state of a changed object during transaction commit,
      the data manager sets the object's _p_changed to a non-None false
      value.

    - Changed -> Ghost

      This transition occurs when a transaction is aborted.  All changed
      objects are invalidated by the data manager by an abort.

    - Ghost -> Saved

      This transition occurs when an attribute or operation of a ghost
      is accessed and the object's state is loaded from the database.

    Note that there is a separate C API that is not included here.
    The C API requires a specific data layout and defines the sticky
    state.


    About Invalidation, Deactivation and the Sticky & Ghost States

    The sticky state is intended to be a short-lived state, to prevent
    an object's state from being discarded while we're in C routines.  It
    is an error to invalidate an object in the sticky state.

    Deactivation is a request that an object discard its state (become
    a ghost).  Deactivation is an optimization, and a request to
    deactivate may be ignored.  There are two equivalent ways to
    request deactivation:

          - call _p_deactivate()
          - set _p_changed to None

    There are two ways to invalidate an object:  call the
    _p_invalidate() method (preferred) or delete its _p_changed
    attribute.  This cannot be ignored, and is used when semantics
    require invalidation.  Normally, an invalidated object transitions
    to the ghost state.  However, some objects cannot be ghosts.  When
    these objects are invalidated, they immediately reload their state
    from their data manager, and are then in the saved state.

    reprs

    By default, persistent objects include the reprs of their
    _p_oid and _p_jar, if any, in their repr. If a subclass implements
    the optional method ``_p_repr``, it will be called and its results returned
    instead of the default repr; if this method raises an exception, that
    exception will be caught and its repr included in the default repr.

    """

    _p_jar = Attribute(
        """The data manager for the object.

        The data manager should implement IPersistentDataManager (note that
        this constraint is not enforced).

        If there is no data manager, then this is None.

        Once assigned to a data manager, an object cannot be re-assigned
        to another.
        """)

    _p_oid = Attribute(
        """The object id.

        It is up to the data manager to assign this.

        The special value None is reserved to indicate that an object
        id has not been assigned.  Non-None object ids must be non-empty
        strings.  The 8-byte string consisting of 8 NUL bytes
        ('\x00\x00\x00\x00\x00\x00\x00\x00') is reserved to identify the
        database root object.

        Once assigned an OID, an object cannot be re-assigned another.
        """)

    _p_changed = Attribute(
        """The persistent state of the object.

        This is one of:

        None -- The object is a ghost.

        false but not None -- The object is saved (or has never been saved).

        true -- The object has been modified since it was last saved.

        The object state may be changed by assigning or deleting this
        attribute; however, assigning None is ignored if the object is
        not in the saved state, and may be ignored even if the object is
        in the saved state.

        At and after ZODB 3.6, setting _p_changed to a true value for a ghost
        object activates the object; prior to 3.6, setting _p_changed to a
        true value on a ghost object was ignored.

        Note that an object can transition to the changed state only if
        it has a data manager.  When such a state change occurs, the
        'register' method of the data manager must be called, passing the
        persistent object.

        Deleting this attribute forces invalidation independent of
        existing state, although it is an error if the sticky state is
        current.
        """)

    _p_serial = Attribute(
        """The object serial number.

        This member is used by the data manager to distiguish distinct
        revisions of a given persistent object.

        This is an 8-byte string (not Unicode).
        """)

    _p_mtime = Attribute(
        """The object's modification time (read-only).

        This is a float, representing seconds since the epoch (as returned
        by time.time).
        """)

    _p_state = Attribute(
        """The object's persistence state token.

        Must be one of GHOST, UPTODATE, CHANGED, or STICKY.
        """)

    _p_estimated_size = Attribute(
        """An estimate of the object's size in bytes.

        May be set by the data manager.
        """)

    # Attribute access protocol
    def __getattribute__(name):
        """ Handle activating ghosts before returning an attribute value.

        "Special" attributes and '_p_*' attributes don't require activation.
        """

    def __setattr__(name, value):
        """ Handle activating ghosts before setting an attribute value.

        "Special" attributes and '_p_*' attributes don't require activation.
        """

    def __delattr__(name):
        """ Handle activating ghosts before deleting an attribute value.

        "Special" attributes and '_p_*' attributes don't require activation.
        """

    # Pickling protocol.
    def __getstate__():
        """Get the object data.

        The state should not include persistent attributes ("_p_name").
        The result must be picklable.
        """

    def __setstate__(state):
        """Set the object data.
        """

    def __reduce__():
        """Reduce an object to contituent parts for serialization.
        """

    # Custom methods
    def _p_activate():
        """Activate the object.

        Change the object to the saved state if it is a ghost.
        """

    def _p_deactivate():
        """Deactivate the object.

        Possibly change an object in the saved state to the
        ghost state.  It may not be possible to make some persistent
        objects ghosts, and, for optimization reasons, the implementation
        may choose to keep an object in the saved state.
        """

    def _p_invalidate():
        """Invalidate the object.

        Invalidate the object.  This causes any data to be thrown
        away, even if the object is in the changed state.  The object
        is moved to the ghost state; further accesses will cause
        object data to be reloaded.
        """

    def _p_getattr(name):
        """Test whether the base class must handle the name

        The method unghostifies the object, if necessary.
        The method records the object access, if necessary.

        This method should be called by subclass __getattribute__
        implementations before doing anything else. If the method
        returns True, then __getattribute__ implementations must delegate
        to the base class, Persistent.
        """

    def _p_setattr(name, value):
        """Save persistent meta data

        This method should be called by subclass __setattr__ implementations
        before doing anything else.  If it returns true, then the attribute
        was handled by the base class.

        The method unghostifies the object, if necessary.
        The method records the object access, if necessary.
        """

    def _p_delattr(name):
        """Delete persistent meta data

        This method should be called by subclass __delattr__ implementations
        before doing anything else.  If it returns true, then the attribute
        was handled by the base class.

        The method unghostifies the object, if necessary.
        The method records the object access, if necessary.
        """

# TODO:  document conflict resolution.

class IPersistentDataManager(Interface):
    """Provide services for managing persistent state.

    This interface is used by a persistent object to interact with its
    data manager in the context of a transaction.
    """
    _cache = Attribute("The pickle cache associated with this connection.")

    def setstate(object):
        """Load the state for the given object.

        The object should be in the ghost state. The object's state will be
        set and the object will end up in the saved state.

        The object must provide the IPersistent interface.
        """

    def oldstate(obj, tid):
        """Return copy of 'obj' that was written by transaction 'tid'.

        The returned object does not have the typical metadata (_p_jar, _p_oid,
        _p_serial) set. I'm not sure how references to other peristent objects
        are handled.

        Parameters
        obj: a persistent object from this Connection.
        tid: id of a transaction that wrote an earlier revision.

        Raises KeyError if tid does not exist or if tid deleted a revision of
        obj.
        """

    def register(object):
        """Register an IPersistent with the current transaction.

        This method must be called when the object transitions to
        the changed state.

        A subclass could override this method to customize the default
        policy of one transaction manager for each thread.
        """

# Maybe later:
##     def mtime(object):
##         """Return the modification time of the object.

##         The modification time may not be known, in which case None
##         is returned.  If non-None, the return value is the kind of
##         timestamp supplied by Python's time.time().
##         """


class IPickleCache(Interface):
    """ API of the cache for a ZODB connection.
    """
    def __getitem__(oid):
        """ -> the persistent object for OID.

        o Raise KeyError if not found.
        """

    def __setitem__(oid, value):
        """ Save the persistent object under OID.

        o 'oid' must be a string, else raise ValueError.

        o Raise KeyError on duplicate
        """

    def __delitem__(oid):
        """ Remove the persistent object for OID.

        o 'oid' must be a string, else raise ValueError.

        o Raise KeyError if not found.
        """

    def get(oid, default=None):
        """ -> the persistent object for OID.

        o Return 'default' if not found.
        """

    def __len__():
        """ -> the number of OIDs in the cache.
        """

    def items():
        """-> a sequence of tuples (oid, value) for cached objects.

        o Only includes items in 'data' (no p-classes).
        """

    def ringlen():
        """ -> the number of persistent objects in the ring.

        o Only includes items in the ring (no ghosts or p-classes).
        """

    def lru_items():
        """ -> a sequence of tuples (oid, value) for cached objects.

        o Tuples will be in LRU order.

        o Only includes items in the ring (no ghosts or p-classes).
        """

    def klass_items():
        """-> a sequence of tuples (oid, value) for cached p-classes.

        o Only includes persistent classes.
        """

    def incrgc():
        """ Perform an incremental garbage collection sweep.

        o Reduce number of non-ghosts to 'cache_size', if possible.

        o Ghostify in LRU order.

        o Skip dirty or sticky objects.

        o Quit once we get down to 'cache_size'.
        """

    def full_sweep():
        """ Perform a full garbage collection sweep.

        o Reduce number of non-ghosts to 0, if possible.

        o Ghostify all non-sticky / non-changed objecs.
        """

    def minimize():
        """ Alias for 'full_sweep'.

        o XXX?
        """

    def new_ghost(oid, obj):
        """ Add the given (ghost) object to the cache.

        Also, set its _p_jar and _p_oid, and ensure it is in the
        GHOST state.

        If the object doesn't define '_p_oid' / '_p_jar', raise.

        If the object's '_p_oid' is not None, raise.

        If the object's '_p_jar' is not None, raise.

        If 'oid' is already in the cache, raise.
        """


    def invalidate(to_invalidate):
        """ Invalidate the indicated objects.

        o If 'to_invalidate' is a string, treat it as an OID.

        o Otherwise, iterate over it as a sequence of OIDs.

        o Any OID corresponding to a p-class will cause the corresponding
            p-class to be removed from the cache.

        o For all other OIDs, ghostify the corrsponding object and
            remove it from the ring.
        """

    def debug_info():
        """Return debugging data about objects in the cache.

        o Return a sequence of tuples, (oid, refcount, typename, state).
        """

    def update_object_size_estimation(oid, new_size):
        """Update the cache's size estimation for 'oid', if known to the cache.
        """

    cache_size = Attribute('Target size of the cache')
    cache_drain_resistance = Attribute('Factor for draining cache below '
                                        'target size')
    cache_non_ghost_count = Attribute('Number of non-ghosts in the cache '
                                        '(XXX how is it different from '
                                        'ringlen?')
    cache_data = Attribute("Property:  copy of our 'data' dict")
    cache_klass_count = Attribute("Property: len of 'persistent_classes'")


class IExtendedPickleCache(IPickleCache):
    """
    Extra operations for a pickle cache.
    """

    def reify(to_reify):
        """ Reify the indicated objects.

        o If 'to_reify' is a string, treat it as an OID.

        o Otherwise, iterate over it as a sequence of OIDs.

        o For each OID, if present in 'data' and in GHOST state:

            o Call '_p_activate' on the object.

            o Add it to the ring.

        o If any OID is present but not in GHOST state, skip it.

        o Raise KeyErrory if any OID is not present.
        """

    def mru(oid):
        """ Move the element corresonding to 'oid' to the head.

        o Raise KeyError if no element is found.
        """



print("import: 'persistent'")
import persistent

print("import: 'persistent._timestamp'")
import persistent._timestamp

print("import: 'persistent.cPersistence'")
import persistent.cPersistence

print("import: 'persistent.cPickleCache'")
import persistent.cPickleCache




##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import gc
import unittest

from persistent.interfaces import UPTODATE
from persistent._compat import PYPY
from persistent.tests.utils import skipIfNoCExtension


# pylint:disable=protected-access,too-many-lines,too-many-public-methods
# pylint:disable=attribute-defined-outside-init,redefined-outer-name

_marker = object()

class DummyPersistent(object):
    _Persistent__ring = None

    def _p_invalidate(self):
        from persistent.interfaces import GHOST
        self._p_state = GHOST

    _p_deactivate = _p_invalidate

    def _p_invalidate_deactivate_helper(self, clear=True):
        self._p_invalidate()

    def _p_activate(self):
        self._p_state = UPTODATE


class DummyConnection(object):
    pass


def _len(seq):
    return len(list(seq))


class PickleCacheTestMixin(object):

    # py2/3 compat
    assertRaisesRegex = getattr(unittest.TestCase,
                                'assertRaisesRegex',
                                unittest.TestCase.assertRaisesRegexp)

    def _getTargetClass(self):
        from persistent.picklecache import PickleCachePy as BasePickleCache
        class PickleCache(BasePickleCache):
            _CACHEABLE_TYPES = BasePickleCache._CACHEABLE_TYPES + (DummyPersistent,)
        return PickleCache

    def _getTargetInterface(self):
        from persistent.interfaces import IPickleCache
        return IPickleCache

    def _makeOne(self, jar=None, target_size=10):
        if jar is None:
            jar = DummyConnection()
        return self._getTargetClass()(jar, target_size)

    def _getDummyPersistentClass(self):
        return DummyPersistent

    def _getRealPersistentClass(self):
        from persistent.persistence import PersistentPy
        return PersistentPy

    def _makePersist(self, state=None, oid=b'foo', jar=_marker, kind=_marker):
        from persistent.interfaces import GHOST

        if state is None:
            state = GHOST
        if jar is _marker:
            jar = DummyConnection()
        kind = self._getDummyPersistentClass() if kind is _marker else kind
        persist = kind()
        try:
            persist._p_state = state
        except AttributeError:
            pass
        persist._p_oid = oid
        persist._p_jar = jar
        return persist

    def test_class_conforms_to_IPickleCache(self):
        from zope.interface.verify import verifyClass
        verifyClass(self._getTargetInterface(), self._getTargetClass())

    def test_instance_conforms_to_IPickleCache(self):
        from zope.interface.verify import verifyObject
        verifyObject(self._getTargetInterface(), self._makeOne())

    def test_empty(self):
        cache = self._makeOne()

        self.assertEqual(len(cache), 0)
        self.assertEqual(_len(cache.items()), 0)
        self.assertEqual(_len(cache.klass_items()), 0)
        self.assertEqual(cache.ringlen(), 0)
        self.assertEqual(len(cache.lru_items()), 0)
        self.assertEqual(cache.cache_size, 10)
        self.assertEqual(cache.cache_drain_resistance, 0)
        self.assertEqual(cache.cache_non_ghost_count, 0)
        self.assertEqual(dict(cache.cache_data), {})
        self.assertEqual(cache.cache_klass_count, 0)

    def test___getitem___nonesuch_raises_KeyError(self):
        cache = self._makeOne()

        self.assertRaises(KeyError, lambda: cache['nonesuch'])

    def test_get_nonesuch_no_default(self):
        cache = self._makeOne()

        self.assertEqual(cache.get('nonesuch'), None)

    def test_get_nonesuch_w_default(self):
        cache = self._makeOne()
        default = object

        self.assertIs(cache.get('nonesuch', default), default)

    def test___setitem___non_string_oid_raises_TypeError(self):
        cache = self._makeOne()

        with self.assertRaises(TypeError):
            cache[object()] = self._makePersist()

    def test___setitem___duplicate_oid_same_obj(self):

        KEY = b'original'
        cache = self._makeOne()
        original = self._makePersist(oid=KEY)
        cache[KEY] = original
        cache[KEY] = original

    def test___setitem___duplicate_oid_raises_ValueError(self):

        KEY = b'original'
        cache = self._makeOne()
        original = self._makePersist(oid=KEY)
        cache[KEY] = original
        duplicate = self._makePersist(oid=KEY)

        with self.assertRaises(ValueError):
            cache[KEY] = duplicate

    def test___setitem___ghost(self):
        from persistent.interfaces import GHOST

        KEY = b'ghost'
        cache = self._makeOne()
        ghost = self._makePersist(state=GHOST, oid=KEY)

        cache[KEY] = ghost

        self.assertEqual(len(cache), 1)
        items = list(cache.items())
        self.assertEqual(len(items), 1)
        self.assertEqual(_len(cache.klass_items()), 0)
        self.assertEqual(items[0][0], KEY)
        self.assertIs(items[0][1], ghost)
        self.assertIs(cache[KEY], ghost)
        return cache

    def test___setitem___mismatch_key_oid(self):
        KEY = b'uptodate'
        cache = self._makeOne()
        uptodate = self._makePersist(state=UPTODATE)

        with self.assertRaises(ValueError):
            cache[KEY] = uptodate


    def test___setitem___non_ghost(self):
        KEY = b'uptodate'
        cache = self._makeOne()
        uptodate = self._makePersist(state=UPTODATE, oid=KEY)

        cache[KEY] = uptodate

        self.assertEqual(len(cache), 1)
        items = list(cache.items())
        self.assertEqual(len(items), 1)
        self.assertEqual(_len(cache.klass_items()), 0)
        self.assertEqual(items[0][0], KEY)
        self.assertEqual(cache.ringlen(), 1)
        self.assertTrue(items[0][1] is uptodate)
        self.assertTrue(cache[KEY] is uptodate)
        self.assertTrue(cache.get(KEY) is uptodate)

    def test___setitem___persistent_class(self):

        KEY = b'pclass'
        class pclass(object):
            _p_oid = KEY
            _p_jar = DummyConnection()
        cache = self._makeOne(pclass._p_jar)

        cache[KEY] = pclass

        kitems = list(cache.klass_items())
        self.assertEqual(len(cache), 1)
        self.assertEqual(len(kitems), 1)
        self.assertEqual(kitems[0][0], KEY)
        self.assertIs(kitems[0][1], pclass)
        self.assertIs(cache[KEY], pclass)
        self.assertIs(cache.get(KEY), pclass)
        return cache

    def test___delitem___non_string_oid_raises_TypeError(self):
        cache = self._makeOne()

        with self.assertRaises(TypeError):
            del cache[object()]

    def test___delitem___nonesuch_raises_KeyError(self):

        cache = self._makeOne()

        with self.assertRaises(KeyError):
            del cache[b'nonesuch']

    def test___delitem___w_persistent_class(self):

        KEY = b'pclass'
        cache = self._makeOne()
        class pclass(object):
            _p_oid = KEY
            _p_jar = DummyConnection()
        cache = self._makeOne()

        cache[KEY] = pclass
        del cache[KEY]
        self.assertIs(cache.get(KEY, self), self)
        self.assertEqual(cache.ringlen(), 0)
        return cache, KEY

    def test___delitem___w_normal_object(self):
        KEY = b'uptodate'
        cache = self._makeOne()
        uptodate = self._makePersist(state=UPTODATE, oid=KEY)

        cache[KEY] = uptodate

        del cache[KEY]
        self.assertTrue(cache.get(KEY, self) is self)

    def test___delitem___w_ghost(self):
        from persistent.interfaces import GHOST

        cache = self._makeOne()
        KEY = b'ghost'
        ghost = self._makePersist(state=GHOST, oid=KEY)

        cache[KEY] = ghost

        del cache[KEY]
        self.assertTrue(cache.get(KEY, self) is self)

    def test___delitem___w_remaining_object(self):
        cache = self._makeOne()
        REMAINS = b'remains'
        UPTODATE = b'uptodate'
        remains = self._makePersist(state=UPTODATE, oid=REMAINS)
        uptodate = self._makePersist(state=UPTODATE, oid=UPTODATE)

        cache[REMAINS] = remains
        cache[UPTODATE] = uptodate

        del cache[UPTODATE]
        self.assertTrue(cache.get(UPTODATE, self) is self)
        self.assertTrue(cache.get(REMAINS, self) is remains)

    def test_lruitems(self):
        cache = self._makeOne()
        ONE = b'one'
        TWO = b'two'
        THREE = b'three'
        cache[ONE] = self._makePersist(oid=b'one', state=UPTODATE)
        cache[TWO] = self._makePersist(oid=b'two', state=UPTODATE)
        cache[THREE] = self._makePersist(oid=b'three', state=UPTODATE)

        items = cache.lru_items()
        self.assertEqual(_len(items), 3)
        self.assertEqual(items[0][0], ONE)
        self.assertEqual(items[1][0], TWO)
        self.assertEqual(items[2][0], THREE)


    def _numbered_oid(self, i):
        # Python 3.4 doesn't support % on bytes,
        # so we go the long way
        oid_s = 'oid_%04d' % i
        return oid_s.encode('ascii')

    def _populate_cache(self, cache, count=100,
                        state_0=UPTODATE,
                        state_rest=UPTODATE):

        oids = []
        for i in range(100):
            oid = self._numbered_oid(i)
            oids.append(oid)
            state = state_0 if i == 0 else state_rest
            cache[oid] = self._makePersist(oid=oid, state=state)
        return oids

    def test_incrgc_simple(self):
        cache = self._makeOne()
        oids = self._populate_cache(cache)
        self.assertEqual(cache.cache_non_ghost_count, 100)

        cache.incrgc()
        gc.collect() # banish the ghosts who are no longer in the ring

        self.assertEqual(cache.cache_non_ghost_count, 10)
        items = cache.lru_items()
        self.assertEqual(_len(items), 10)
        self.assertEqual(items[0][0], b'oid_0090')
        self.assertEqual(items[1][0], b'oid_0091')
        self.assertEqual(items[2][0], b'oid_0092')
        self.assertEqual(items[3][0], b'oid_0093')
        self.assertEqual(items[4][0], b'oid_0094')
        self.assertEqual(items[5][0], b'oid_0095')
        self.assertEqual(items[6][0], b'oid_0096')
        self.assertEqual(items[7][0], b'oid_0097')
        self.assertEqual(items[8][0], b'oid_0098')
        self.assertEqual(items[9][0], b'oid_0099')

        for oid in oids[:90]:
            self.assertIsNone(cache.get(oid))

        for oid in oids[90:]:
            self.assertIsNotNone(cache.get(oid))

    def test_incrgc_w_smaller_drain_resistance(self):
        cache = self._makeOne()
        cache.cache_drain_resistance = 2
        self._populate_cache(cache)
        self.assertEqual(cache.cache_non_ghost_count, 100)

        cache.incrgc()

        self.assertEqual(cache.cache_non_ghost_count, 10)

    def test_incrgc_w_larger_drain_resistance(self):
        cache = self._makeOne()
        cache.cache_drain_resistance = 2
        cache.cache_size = 90
        self._populate_cache(cache)

        self.assertEqual(cache.cache_non_ghost_count, 100)

        cache.incrgc()

        self.assertEqual(cache.cache_non_ghost_count, 49)

    def test_full_sweep(self):
        cache = self._makeOne()
        oids = self._populate_cache(cache)
        self.assertEqual(cache.cache_non_ghost_count, 100)

        cache.full_sweep()
        gc.collect() # banish the ghosts who are no longer in the ring

        self.assertEqual(cache.cache_non_ghost_count, 0)

        for oid in oids:
            self.assertTrue(cache.get(oid) is None)


    def test_minimize(self):
        cache = self._makeOne()
        oids = self._populate_cache(cache)
        self.assertEqual(cache.cache_non_ghost_count, 100)

        cache.minimize()
        gc.collect() # banish the ghosts who are no longer in the ring

        self.assertEqual(cache.cache_non_ghost_count, 0)

        for oid in oids:
            self.assertTrue(cache.get(oid) is None)

    def test_minimize_turns_into_ghosts(self):
        from persistent.interfaces import GHOST

        cache = self._makeOne()
        oid = self._numbered_oid(1)
        obj = cache[oid] = self._makePersist(oid=oid, state=UPTODATE)
        self.assertEqual(cache.cache_non_ghost_count, 1)

        cache.minimize()
        gc.collect() # banish the ghosts who are no longer in the ring

        self.assertEqual(cache.cache_non_ghost_count, 0)

        self.assertEqual(obj._p_state, GHOST)

    def test_new_ghost_non_persistent_object(self):

        cache = self._makeOne()
        with self.assertRaises((AttributeError, TypeError)):
            cache.new_ghost(b'123', object())

    def test_new_ghost_obj_already_has_oid(self):

        from persistent.interfaces import GHOST
        candidate = self._makePersist(oid=b'123', state=GHOST)
        cache = self._makeOne()
        with self.assertRaises(ValueError):
            cache.new_ghost(b'123', candidate)

    def test_new_ghost_obj_already_has_jar(self):
        cache = self._makeOne()
        candidate = self._makePersist(oid=None, jar=object())
        with self.assertRaises(ValueError):
            cache.new_ghost(b'123', candidate)

    def test_new_ghost_obj_already_in_cache(self):
        KEY = b'123'
        cache = self._makeOne()
        candidate = self._makePersist(oid=KEY)
        cache[KEY] = candidate
        # Now, normally we can't get in the cache without an oid and jar
        # (the C implementation doesn't allow it), so if we try to create
        # a ghost, we get the value error
        self.assertRaises(ValueError, cache.new_ghost, KEY, candidate)
        return cache, KEY, candidate

    def test_new_ghost_success_already_ghost(self):
        from persistent.interfaces import GHOST

        KEY = b'123'
        jar = DummyConnection()
        cache = self._makeOne(jar)
        candidate = self._makePersist(oid=None, jar=None)
        cache.new_ghost(KEY, candidate)
        self.assertTrue(cache.get(KEY) is candidate)
        self.assertEqual(candidate._p_oid, KEY)
        self.assertEqual(candidate._p_jar, jar)
        self.assertEqual(candidate._p_state, GHOST)

    def test_new_ghost_success_not_already_ghost(self):
        from persistent.interfaces import GHOST

        KEY = b'123'
        jar = DummyConnection()
        cache = self._makeOne(jar)
        candidate = self._makePersist(oid=None, jar=None, state=UPTODATE)
        cache.new_ghost(KEY, candidate)
        self.assertTrue(cache.get(KEY) is candidate)
        self.assertEqual(candidate._p_oid, KEY)
        self.assertEqual(candidate._p_jar, jar)
        self.assertEqual(candidate._p_state, GHOST)

    def test_new_ghost_w_pclass_non_ghost(self):
        KEY = b'123'
        class Pclass(object):
            _p_oid = None
            _p_jar = None
        cache = self._makeOne()
        cache.new_ghost(KEY, Pclass)
        self.assertTrue(cache.get(KEY) is Pclass)
        self.assertEqual(Pclass._p_oid, KEY)
        return cache, Pclass, KEY

    def test_new_ghost_w_pclass_ghost(self):
        KEY = b'123'
        class Pclass(object):
            _p_oid = None
            _p_jar = None
        cache = self._makeOne()
        cache.new_ghost(KEY, Pclass)
        self.assertTrue(cache.get(KEY) is Pclass)
        self.assertEqual(Pclass._p_oid, KEY)
        return cache, Pclass, KEY

    def test_invalidate_miss_single(self):
        KEY = b'123'
        cache = self._makeOne()
        cache.invalidate(KEY) # doesn't raise

    def test_invalidate_miss_multiple(self):
        KEY = b'123'
        KEY2 = b'456'
        cache = self._makeOne()
        cache.invalidate([KEY, KEY2]) # doesn't raise

    def test_invalidate_hit_single_non_ghost(self):
        from persistent.interfaces import GHOST

        KEY = b'123'
        jar = DummyConnection()
        cache = self._makeOne(jar)
        candidate = self._makePersist(oid=b'123', jar=jar, state=UPTODATE)
        cache[KEY] = candidate
        self.assertEqual(cache.ringlen(), 1)
        cache.invalidate(KEY)
        self.assertEqual(cache.ringlen(), 0)
        self.assertEqual(candidate._p_state, GHOST)

    def test_invalidate_hit_multiple_non_ghost(self):
        from persistent.interfaces import GHOST

        KEY = b'123'
        KEY2 = b'456'
        jar = DummyConnection()
        cache = self._makeOne()
        c1 = self._makePersist(oid=KEY, jar=jar, state=UPTODATE)
        cache[KEY] = c1
        c2 = self._makePersist(oid=KEY2, jar=jar, state=UPTODATE)
        cache[KEY2] = c2
        self.assertEqual(cache.ringlen(), 2)
        # These should be in the opposite order of how they were
        # added to the ring to ensure ring traversal works
        cache.invalidate([KEY2, KEY])
        self.assertEqual(cache.ringlen(), 0)
        self.assertEqual(c1._p_state, GHOST)
        self.assertEqual(c2._p_state, GHOST)

    def test_debug_info_w_persistent_class(self):
        KEY = b'pclass'
        class pclass(object):
            _p_oid = KEY
            _p_jar = DummyConnection()
        cache = self._makeOne(pclass._p_jar)
        pclass._p_state = UPTODATE
        cache[KEY] = pclass

        gc.collect() # pypy vs. refcounting
        info = cache.debug_info()

        self.assertEqual(len(info), 1)
        # C and Python return different length tuples,
        # and the refcounts are off by one.
        oid = info[0][0]
        typ = info[0][2]

        self.assertEqual(oid, KEY)
        self.assertEqual(typ, 'type')

        return pclass, info[0]

    def test_debug_info_w_normal_object(self):
        KEY = b'uptodate'
        cache = self._makeOne()
        uptodate = self._makePersist(state=UPTODATE, oid=KEY)
        cache[KEY] = uptodate

        gc.collect() # pypy vs. refcounting
        info = cache.debug_info()

        self.assertEqual(len(info), 1)
        # C and Python return different length tuples,
        # and the refcounts are off by one.
        oid = info[0][0]
        typ = info[0][2]
        self.assertEqual(oid, KEY)
        self.assertEqual(typ, type(uptodate).__name__)
        return uptodate, info[0]


    def test_debug_info_w_ghost(self):
        from persistent.interfaces import GHOST

        KEY = b'ghost'
        cache = self._makeOne()
        ghost = self._makePersist(state=GHOST, oid=KEY)
        cache[KEY] = ghost

        gc.collect() # pypy vs. refcounting
        info = cache.debug_info()

        self.assertEqual(len(info), 1)
        oid, _refc, typ, state = info[0]
        self.assertEqual(oid, KEY)
        self.assertEqual(typ, type(ghost).__name__)
        # In the C implementation, we couldn't actually set the _p_state
        # directly.
        self.assertEqual(state, ghost._p_state)
        return ghost, info[0]

    def test_setting_non_persistent_item(self):
        cache = self._makeOne()
        with self.assertRaisesRegex(TypeError,
                                    "Cache values must be persistent objects."):
            cache[b'12345678'] = object()

    def test_setting_without_jar(self):
        cache = self._makeOne()
        p = self._makePersist(jar=None)
        with self.assertRaisesRegex(ValueError,
                                    "Cached object jar missing"):
            cache[p._p_oid] = p

    def test_setting_already_cached(self):
        jar = DummyConnection()
        cache1 = self._makeOne(jar)
        p = self._makePersist(jar=jar)

        cache1[p._p_oid] = p

        cache2 = self._makeOne()
        with self.assertRaisesRegex(ValueError,
                                    "Cache values may only be in one cache"):
            cache2[p._p_oid] = p

    def test_cache_size(self):
        size = 42
        cache = self._makeOne(target_size=size)
        self.assertEqual(cache.cache_size, size)

        cache.cache_size = 64
        self.assertEqual(cache.cache_size, 64)

    def test_sweep_empty(self):
        cache = self._makeOne()
        # Python returns 0, C returns None
        self.assertFalse(cache.incrgc())

    def test_invalidate_persistent_class_calls_p_invalidate(self):
        KEY = b'pclass'
        class pclass(object):
            _p_oid = KEY
            _p_jar = DummyConnection()
            invalidated = False
            @classmethod
            def _p_invalidate(cls):
                cls.invalidated = True


        cache = self._makeOne(pclass._p_jar)

        cache[KEY] = pclass

        cache.invalidate(KEY)

        self.assertTrue(pclass.invalidated)

    def test_cache_raw(self):
        raw = self._makePersist(kind=self._getRealPersistentClass())
        cache = self._makeOne(raw._p_jar)

        cache[raw._p_oid] = raw
        self.assertEqual(1, len(cache))
        self.assertIs(cache.get(raw._p_oid), raw)

        del raw
        self.assertEqual(1, len(cache))


class PythonPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
    # Tests that depend on the implementation details of the
    # Python PickleCache and the Python persistent object.
    # Anything that involves directly setting the _p_state of a persistent
    # object has to be here, we can't do that in the C implementation.

    def _getTargetInterface(self):
        from persistent.interfaces import IExtendedPickleCache
        return IExtendedPickleCache

    def test_sweep_of_non_deactivating_object(self):
        jar = DummyConnection()
        cache = self._makeOne(jar)
        p = self._makePersist(jar=jar)

        p._p_state = 0 # non-ghost, get in the ring
        cache[p._p_oid] = p

        def bad_deactivate():
            "Doesn't call super, for it's own reasons, so can't be ejected"


        p._p_deactivate = bad_deactivate

        cache._SWEEPABLE_TYPES = DummyPersistent
        self.assertEqual(cache.full_sweep(), 0)
        del cache._SWEEPABLE_TYPES

        del p._p_deactivate
        self.assertEqual(cache.full_sweep(), 1)

    def test_update_object_size_estimation_simple(self):
        cache = self._makeOne()
        p = self._makePersist(jar=cache.jar)

        cache[p._p_oid] = p
        # The cache accesses the private attribute directly to bypass
        # the bit conversion.
        # Note that the _p_estimated_size is set *after*
        # the update call is made in ZODB's serialize
        p._Persistent__size = 0

        cache.update_object_size_estimation(p._p_oid, 2)

        self.assertEqual(cache.total_estimated_size, 64)

        # A missing object does nothing
        cache.update_object_size_estimation(None, 2)
        self.assertEqual(cache.total_estimated_size, 64)

    def test_reify_miss_single(self):
        KEY = b'123'
        cache = self._makeOne()
        self.assertRaises(KeyError, cache.reify, KEY)

    def test_reify_miss_multiple(self):
        KEY = b'123'
        KEY2 = b'456'
        cache = self._makeOne()
        self.assertRaises(KeyError, cache.reify, [KEY, KEY2])

    def test_reify_hit_single_non_ghost(self):
        KEY = b'123'
        jar = DummyConnection()
        cache = self._makeOne(jar)
        candidate = self._makePersist(oid=KEY, jar=jar, state=UPTODATE)
        cache[KEY] = candidate
        self.assertEqual(cache.ringlen(), 1)
        cache.reify(KEY)
        self.assertEqual(cache.ringlen(), 1)
        self.assertEqual(candidate._p_state, UPTODATE)

    def test_cannot_update_mru_while_already_locked(self):
        cache = self._makeOne()
        cache._is_sweeping_ring = True

        updated = cache.mru(None)
        self.assertFalse(updated)

    def test___delitem___w_persistent_class(self):
        cache, key = super(PythonPickleCacheTests, self).test___delitem___w_persistent_class()
        self.assertNotIn(key, cache.persistent_classes)

    def test___setitem___ghost(self):
        cache = super(PythonPickleCacheTests, self).test___setitem___ghost()
        self.assertEqual(cache.ringlen(), 0)

    def test___setitem___persistent_class(self):
        cache = super(PythonPickleCacheTests, self).test___setitem___persistent_class()
        self.assertEqual(_len(cache.items()), 0)

    def test_new_ghost_w_pclass_non_ghost(self):
        cache, Pclass, key = super(PythonPickleCacheTests, self).test_new_ghost_w_pclass_non_ghost()
        self.assertEqual(Pclass._p_jar, cache.jar)
        self.assertIs(cache.persistent_classes[key], Pclass)

    def test_new_ghost_w_pclass_ghost(self):
        cache, Pclass, key = super(PythonPickleCacheTests, self).test_new_ghost_w_pclass_ghost()
        self.assertEqual(Pclass._p_jar, cache.jar)
        self.assertIs(cache.persistent_classes[key], Pclass)

    def test_mru_nonesuch_raises_KeyError(self):
        cache = self._makeOne()

        self.assertRaises(KeyError, cache.mru, b'nonesuch')

    def test_mru_normal(self):
        ONE = b'one'
        TWO = b'two'
        THREE = b'three'
        cache = self._makeOne()
        cache[ONE] = self._makePersist(oid=b'one', state=UPTODATE)
        cache[TWO] = self._makePersist(oid=b'two', state=UPTODATE)
        cache[THREE] = self._makePersist(oid=b'three', state=UPTODATE)

        cache.mru(TWO)

        self.assertEqual(cache.ringlen(), 3)
        items = cache.lru_items()
        self.assertEqual(_len(items), 3)
        self.assertEqual(items[0][0], ONE)
        self.assertEqual(items[1][0], THREE)
        self.assertEqual(items[2][0], TWO)

    def test_mru_ghost(self):
        from persistent.interfaces import GHOST

        ONE = b'one'
        TWO = b'two'
        THREE = b'three'
        cache = self._makeOne()
        cache[ONE] = self._makePersist(oid=b'one', state=UPTODATE)
        two = cache[TWO] = self._makePersist(oid=b'two', state=GHOST)
        # two must live to survive gc
        self.assertIsNotNone(two)
        cache[THREE] = self._makePersist(oid=b'three', state=UPTODATE)

        cache.mru(TWO)

        self.assertEqual(cache.ringlen(), 2)
        items = cache.lru_items()
        self.assertEqual(_len(items), 2)
        self.assertEqual(items[0][0], ONE)
        self.assertEqual(items[1][0], THREE)

    def test_mru_was_ghost_now_active(self):
        from persistent.interfaces import GHOST

        ONE = b'one'
        TWO = b'two'
        THREE = b'three'
        cache = self._makeOne()
        cache[ONE] = self._makePersist(oid=b'one', state=UPTODATE)
        two = cache[TWO] = self._makePersist(oid=b'two', state=GHOST)
        cache[THREE] = self._makePersist(oid=b'three', state=UPTODATE)

        two._p_state = UPTODATE
        cache.mru(TWO)

        self.assertEqual(cache.ringlen(), 3)
        items = cache.lru_items()
        self.assertEqual(_len(items), 3)
        self.assertEqual(items[0][0], ONE)
        self.assertEqual(items[1][0], THREE)
        self.assertEqual(items[2][0], TWO)

    def test_mru_first(self):
        ONE = b'one'
        TWO = b'two'
        THREE = b'three'
        cache = self._makeOne()
        cache[ONE] = self._makePersist(oid=b'one', state=UPTODATE)
        cache[TWO] = self._makePersist(oid=b'two', state=UPTODATE)
        cache[THREE] = self._makePersist(oid=b'three', state=UPTODATE)

        cache.mru(ONE)

        self.assertEqual(cache.ringlen(), 3)
        items = cache.lru_items()
        self.assertEqual(_len(items), 3)
        self.assertEqual(items[0][0], TWO)
        self.assertEqual(items[1][0], THREE)
        self.assertEqual(items[2][0], ONE)

    def test_mru_last(self):
        ONE = b'one'
        TWO = b'two'
        THREE = b'three'
        cache = self._makeOne()
        cache[ONE] = self._makePersist(oid=b'one', state=UPTODATE)
        cache[TWO] = self._makePersist(oid=b'two', state=UPTODATE)
        cache[THREE] = self._makePersist(oid=b'three', state=UPTODATE)

        cache.mru(THREE)

        self.assertEqual(cache.ringlen(), 3)
        items = cache.lru_items()
        self.assertEqual(_len(items), 3)
        self.assertEqual(items[0][0], ONE)
        self.assertEqual(items[1][0], TWO)
        self.assertEqual(items[2][0], THREE)

    def test_invalidate_hit_pclass(self):
        KEY = b'123'
        class Pclass(object):
            _p_oid = KEY
            _p_jar = DummyConnection()
        cache = self._makeOne(Pclass._p_jar)
        cache[KEY] = Pclass
        self.assertIs(cache.persistent_classes[KEY], Pclass)
        cache.invalidate(KEY)
        self.assertNotIn(KEY, cache.persistent_classes)

    def test_debug_info_w_normal_object(self):
        obj, info = super(PythonPickleCacheTests, self).test_debug_info_w_normal_object()
        self.assertEqual(info[1], len(gc.get_referents(obj)))
        self.assertEqual(info[3], UPTODATE)

    def test_debug_info_w_ghost(self):
        ghost, info = super(PythonPickleCacheTests, self).test_debug_info_w_ghost()
        self.assertEqual(info[1], len(gc.get_referents(ghost)))

    def test_debug_info_w_persistent_class(self):
        pclass, info = super(PythonPickleCacheTests, self).test_debug_info_w_persistent_class()
        self.assertEqual(info[3], UPTODATE)
        self.assertEqual(info[1], len(gc.get_referents(pclass)))

    def test_full_sweep_w_sticky(self):
        from persistent.interfaces import STICKY

        cache = self._makeOne()
        oids = self._populate_cache(cache, state_0=STICKY)
        self.assertEqual(cache.cache_non_ghost_count, 100)

        cache.full_sweep()
        gc.collect() # banish the ghosts who are no longer in the ring

        self.assertEqual(cache.cache_non_ghost_count, 1)

        self.assertTrue(cache.get(oids[0]) is not None)
        for oid in oids[1:]:
            self.assertTrue(cache.get(oid) is None)

    def test_full_sweep_w_changed(self):
        from persistent.interfaces import CHANGED

        cache = self._makeOne()
        oids = self._populate_cache(cache, state_0=CHANGED)
        self.assertEqual(cache.cache_non_ghost_count, 100)

        cache.full_sweep()
        gc.collect() # banish the ghosts who are no longer in the ring

        self.assertEqual(cache.cache_non_ghost_count, 1)

        self.assertTrue(cache.get(oids[0]) is not None)
        for oid in oids[1:]:
            self.assertTrue(cache.get(oid) is None)

    def test_init_with_cacheless_jar(self):
        # Sometimes ZODB tests pass objects that don't
        # have a _cache
        class Jar(object):
            was_set = False
            def __setattr__(self, name, value):
                if name == '_cache':
                    object.__setattr__(self, 'was_set', True)
                raise AttributeError(name)

        jar = Jar()
        self._makeOne(jar)
        self.assertTrue(jar.was_set)

    def test_invalidate_hit_multiple_mixed(self):
        from persistent.interfaces import GHOST

        KEY = b'123'
        KEY2 = b'456'
        jar = DummyConnection()
        cache = self._makeOne()
        c1 = self._makePersist(oid=KEY, jar=jar, state=GHOST)
        cache[KEY] = c1
        c2 = self._makePersist(oid=KEY2, jar=jar, state=UPTODATE)
        cache[KEY2] = c2
        self.assertEqual(cache.ringlen(), 1)
        cache.invalidate([KEY, KEY2])
        self.assertEqual(cache.ringlen(), 0)
        self.assertEqual(c1._p_state, GHOST)
        self.assertEqual(c2._p_state, GHOST)

    def test_invalidate_hit_single_ghost(self):
        from persistent.interfaces import GHOST

        KEY = b'123'
        jar = DummyConnection()
        cache = self._makeOne(jar)
        candidate = self._makePersist(oid=b'123', jar=jar, state=GHOST)
        cache[KEY] = candidate
        self.assertEqual(cache.ringlen(), 0)
        cache.invalidate(KEY)
        self.assertEqual(cache.ringlen(), 0)
        self.assertEqual(candidate._p_state, GHOST)

    def test_reify_hit_multiple_mixed(self):
        from persistent.interfaces import GHOST

        KEY = b'123'
        KEY2 = b'456'
        jar = DummyConnection()
        cache = self._makeOne(jar)
        c1 = self._makePersist(oid=KEY, jar=jar, state=GHOST)
        cache[KEY] = c1
        c2 = self._makePersist(oid=KEY2, jar=jar, state=UPTODATE)
        cache[KEY2] = c2
        self.assertEqual(cache.ringlen(), 1)
        cache.reify([KEY, KEY2])
        self.assertEqual(cache.ringlen(), 2)
        self.assertEqual(c1._p_state, UPTODATE)
        self.assertEqual(c2._p_state, UPTODATE)

    def test_reify_hit_single_ghost(self):
        from persistent.interfaces import GHOST

        KEY = b'123'
        jar = DummyConnection()
        cache = self._makeOne()
        candidate = self._makePersist(oid=KEY, jar=jar, state=GHOST)
        cache[KEY] = candidate
        self.assertEqual(cache.ringlen(), 0)
        cache.reify(KEY)
        self.assertEqual(cache.ringlen(), 1)
        items = cache.lru_items()
        self.assertEqual(items[0][0], KEY)
        self.assertTrue(items[0][1] is candidate)
        self.assertEqual(candidate._p_state, UPTODATE)

    def test_cache_garbage_collection_bytes_also_deactivates_object(self):

        class MyPersistent(self._getDummyPersistentClass()):
            def _p_deactivate(self):
                # mimic what the real persistent object does to update the cache
                # size; if we don't get deactivated by sweeping, the cache size
                # won't shrink so this also validates that _p_deactivate gets
                # called when ejecting an object.
                cache.update_object_size_estimation(self._p_oid, -1)

        cache = self._makeOne()
        cache.cache_size = 1000
        oids = []
        for i in range(100):
            oid = self._numbered_oid(i)
            oids.append(oid)
            o = cache[oid] = self._makePersist(oid=oid, kind=MyPersistent, state=UPTODATE)

            o._Persistent__size = 0 # must start 0, ZODB sets it AFTER updating the size

            cache.update_object_size_estimation(oid, 64)
            o._Persistent__size = 2

        self.assertEqual(cache.cache_non_ghost_count, 100)

        # A GC at this point does nothing
        cache.incrgc()
        self.assertEqual(cache.cache_non_ghost_count, 100)
        self.assertEqual(len(cache), 100)

        # Now if we set a byte target:

        cache.cache_size_bytes = 1
        # verify the change worked as expected
        self.assertEqual(cache.cache_size_bytes, 1)
        # verify our entrance assumption is fulfilled
        self.assertTrue(cache.cache_size > 100)
        self.assertTrue(cache.total_estimated_size > 1)
        # A gc shrinks the bytes
        cache.incrgc()
        self.assertEqual(cache.total_estimated_size, 0)

        # It also shrank the measured size of the cache,
        # though this may require a GC to be visible.
        if PYPY: # pragma: no cover
            gc.collect()
        self.assertEqual(len(cache), 1)


    def test_new_ghost_obj_already_in_cache(self):
        base_result = super(PythonPickleCacheTests, self).test_new_ghost_obj_already_in_cache()
        cache, key, candidate = base_result
        # If we're sneaky and remove the OID and jar, then we get the duplicate
        # key error. Removing them only works because we're not using a real
        # persistent object.
        candidate._p_oid = None
        self.assertRaises(ValueError, cache.new_ghost, key, candidate)

        candidate._p_jar = None
        self.assertRaises(KeyError, cache.new_ghost, key, candidate)

    def test_cache_garbage_collection_bytes_with_cache_size_0(self):

        class MyPersistent(self._getDummyPersistentClass()):
            def _p_deactivate(self):
                # mimic what the real persistent object does to update
                # the cache size; if we don't get deactivated by
                # sweeping, the cache size won't shrink so this also
                # validates that _p_deactivate gets called when
                # ejecting an object.
                cache.update_object_size_estimation(self._p_oid, -1)

        cache = self._makeOne()
        cache.cache_size = 0
        cache.cache_size_bytes = 400
        oids = []
        for i in range(100):
            oid = self._numbered_oid(i)
            oids.append(oid)
            o = cache[oid] = self._makePersist(oid=oid,
                                               kind=MyPersistent,
                                               state=UPTODATE)
            # must start 0, ZODB sets it AFTER updating the size
            o._Persistent__size = 0
            cache.update_object_size_estimation(oid, 1)
            o._Persistent__size = 1
            del o # leave it only in the cache

        self.assertEqual(cache.cache_non_ghost_count, 100)
        self.assertEqual(cache.total_estimated_size, 64 * 100)

        cache.incrgc()
        gc.collect() # banish the ghosts who are no longer in the ring
        self.assertEqual(cache.total_estimated_size, 64 * 6)
        self.assertEqual(cache.cache_non_ghost_count, 6)
        self.assertEqual(len(cache), 6)

        cache.full_sweep()
        gc.collect() # banish the ghosts who are no longer in the ring
        self.assertEqual(cache.total_estimated_size, 0)
        self.assertEqual(cache.cache_non_ghost_count, 0)
        self.assertEqual(len(cache), 0)


@skipIfNoCExtension
class CPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):

    def _getTargetClass(self):
        from persistent._compat import _c_optimizations_available as get_c
        return get_c()['persistent.picklecache'].PickleCache

    def _getRealPersistentClass(self):
        from persistent._compat import _c_optimizations_available as get_c
        return get_c()['persistent.persistence'].Persistent

    def _getDummyPersistentClass(self):
        class DummyPersistent(self._getRealPersistentClass()):
            __slots__ = ()
        return DummyPersistent

    def test_inst_does_not_conform_to_IExtendedPickleCache(self):
        # Test that ``@use_c_impl`` is only applying the correct
        # interface declaration to the C implementation.
        from persistent.interfaces import IExtendedPickleCache
        from zope.interface.verify import verifyObject
        from zope.interface.exceptions import Invalid
        # We don't claim to implement it.
        self.assertFalse(IExtendedPickleCache.providedBy(self._makeOne()))
        # And we don't even provide everything it asks for.
        # (Exact error depends on version of zope.interface and what we
        # fail to implement. 5.0 probably raises MultipleInvalid).
        with self.assertRaises(Invalid):
            verifyObject(IExtendedPickleCache, self._makeOne(), tentative=True)

    def test___setitem___persistent_class(self):
        cache = super(CPickleCacheTests, self).test___setitem___persistent_class()
        self.assertEqual(_len(cache.items()), 1)

    def test_cache_garbage_collection_bytes_with_cache_size_0(self):

        class DummyConnection(object):
            def register(self, obj):
                pass

        dummy_connection = DummyConnection()
        dummy_connection.register(1) # for coveralls

        def makePersistent(oid):
            persist = self._getDummyPersistentClass()()
            persist._p_oid = oid
            persist._p_jar = dummy_connection
            return persist

        cache = self._getTargetClass()(dummy_connection)
        dummy_connection._cache = cache

        cache.cache_size = 0
        cache.cache_size_bytes = 400

        oids = []
        for i in range(100):
            oid = self._numbered_oid(i)
            oids.append(oid)
            o = cache[oid] = makePersistent(oid)
            cache.update_object_size_estimation(oid, 1)
            o._p_estimated_size = 1
            del o # leave it only in the cache

        self.assertEqual(cache.cache_non_ghost_count, 100)
        self.assertEqual(cache.total_estimated_size, 64 * 100)

        cache.incrgc()
        self.assertEqual(cache.total_estimated_size, 64 * 6)
        self.assertEqual(cache.cache_non_ghost_count, 6)
        self.assertEqual(len(cache), 6)

        cache.full_sweep()
        gc.collect() # banish the ghosts who are no longer in the ring
        self.assertEqual(cache.total_estimated_size, 0)
        self.assertEqual(cache.cache_non_ghost_count, 0)
        self.assertEqual(len(cache), 0)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)

if __name__ == '__main__':
    unittest.main()
